from ._version import __version__
from ._costs import HaHCost, L1Cost
from ._layers import AdaptiveThreshold, DivisiveNormalization2d, Normalize, ImplicitNormalizationConv, HaHBlock, FractionalThreshold, AdaptiveThreshold
from ._extract_features import ActivationExtractor
from ._models import HaHVGG16, VGG16



from typing import Dict, Union, Tuple, Optional
import numpy as np
from math import ceil

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


__all__ = ["HaHCost", "L1Cost"]


class _Cost(nn.Module):
    """
    Base cost class
    Args:
        dim: dimension where reduction is applied (Generally channel dimension)
        reduction: Reduction type
    """

    def __init__(self, dim: int = 1, reduction: str = "mean") -> None:
        super(_Cost, self).__init__()
        self.dim = dim
        self.reduction = reduction


class HaHCost(_Cost):
    """
    Hebbian/Anti-Hebbian Cost class
    Args:
        ratio: Top ratio to be promoted
        demote: Hyperparameter determining how much to emphasize the anti-Hebbian component
        dim: dimension where reduction is applied (Generally channel dimension)
        reduction: Reduction type
    """

    def __init__(self, ratio: float = 0.1, demote: float = 1.0, dim: int = 1, reduction: str = "mean") -> None:
        super(HaHCost, self).__init__(dim, reduction)
        self.ratio = ratio
        self.demote = demote

    def forward(self, input: Tensor) -> Tensor:

        N = input.shape[self.dim]
        K = ceil(self.ratio * N)

        o = F.relu(input)
        o = torch.sort(o, dim=self.dim, descending=True)[0]

        top_K_avg = o[:, :K].mean(dim=self.dim)
        bottom_avg = o[:, K:].mean(dim=self.dim)

        if self.reduction == "mean":
            return torch.mean(top_K_avg-self.demote*bottom_avg)
        elif self.reduction == "sum":
            return torch.sum(top_K_avg-self.demote*bottom_avg)
        else:
            raise NotImplementedError

    def __repr__(self) -> str:
        s = f"HaHCost(ratio={self.ratio}, demote={self.demote}, dim={self.dim}, reduction={self.reduction})"
        return s


class L1Cost(_Cost):
    """
    L1 Cost class
    Args:
        dim: dimension where reduction is applied (Generally channel dimension)
        reduction: Reduction type
    """

    def __init__(self, dim: int = 1, reduction: str = "mean") -> None:
        super(HaHCost, self).__init__(dim, reduction)

    def forward(self, input: Tensor, ratio: float) -> Tensor:
        return torch.mean(torch.sum(torch.abs(input), dim=self.dim))

    def __repr__(self) -> str:
        s = f"L1Cost(dim={self.dim}, reduction={self.reduction})"
        return s



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

version = {}
with open("./hahtorch/_version.py") as fp:
    exec(fp.read(), version)

release_tag = "v_" + str(version['version_info'][0]) + \
    str(version['version_info'][1]) + str(version['version_info'][2])

setup(
    name='hahtorch',
    packages=['hahtorch'],
    version=version['__version__'],
    license='MIT',
    description='Hebbian/Anti-Hebbian Learning for Pytorch',
    long_description_content_type="text/markdown",
    long_description=long_description,
    author='Metehan Cekic',
    author_email='metehancekic@ucsb.edu',
    url='https://github.com/metehancekic/HaH.git',
    download_url='https://github.com/metehancekic/HaH/archive/' + release_tag + '.tar.gz',
    keywords=['Hebbian', 'Anti-Hebbian', 'Pytorch', 'Neuro-inspired'],
    install_requires=[
        'numpy',
        ],
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',      # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',   # Again, pick a license
        'Programming Language :: Python :: 3',  # Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        ],
)



from typing import Dict, Iterable, Callable
from collections import OrderedDict
import numpy as np

import torch
from torch import nn, Tensor
from torch.nn.functional import pad


from collections import OrderedDict

__all__ = ["ActivationExtractor"]


class ActivationExtractor(nn.Module):
    """
    Wrapper extracting a specific layer type outputs and inputs from a DNN
    Args:
        model: DNN model
        layer_type: Layer type of the interested outputs and inputs
    """

    def __init__(self, model: nn.Module, layer_type):
        super().__init__()
        self._model = model
        self.layer_type = layer_type

        self.hook_handles = {}

        self.layer_outputs = OrderedDict()
        self.layer_inputs = OrderedDict()
        self.layers_of_interest = OrderedDict()
        for layer_id, layer in model.named_modules():
            if isinstance(layer, layer_type):
                self.layer_outputs[layer_id] = torch.empty(0)
                self.layer_inputs[layer_id] = torch.empty(0)
                self.layers_of_interest[layer_id] = layer

        for layer_id, layer in model.named_modules():
            if isinstance(layer, layer_type):
                self.hook_handles[layer_id] = layer.register_forward_hook(
                    self.generate_hook_fn(layer_id))

    def generate_hook_fn(self, layer_id: str) -> Callable:
        def fn(_, input, output):
            self.layer_outputs[layer_id] = output
            self.layer_inputs[layer_id] = input[0]
        return fn

    def close(self):
        [hook_handle.remove() for hook_handle in self.hook_handles.values()]

    def forward(self, x):
        return self._model(x)

    def __getattribute__(self, name: str):
        # the last three are used in nn.Module.__setattr__
        if name in ["_model", "layers_of_interest", "layer_outputs", "layer_inputs", "hook_handles", "generate_hook_fn", "close", "__dict__", "_parameters", "_buffers", "_non_persistent_buffers_set"]:
            return object.__getattribute__(self, name)
        else:
            return getattr(self._model, name)



from typing import Union, Tuple

import torch
import torch.nn.functional as F
import torch.nn as nn

from math import floor


class ImplicitNormalizationConv(nn.Conv2d):
    r"""
    torch.Conv2d with implicit weight normalization (L2 normalization)
    """

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        weight_norms = (self.weight**2).sum(dim=(1, 2, 3),
                                            keepdim=True).transpose(0, 1).sqrt()

        conv = super().forward(x)
        return conv/(weight_norms+1e-6)


class Normalize(nn.Module):
    r"""Data normalizing class as torch.nn.Module

    Args:
        mean (float): Mean value of the training dataset.
        std (float): Standard deviation value of the training dataset.

    """

    def __init__(self, mean, std):
        super(Normalize, self).__init__()
        self.mean = torch.Tensor(mean)
        self.std = torch.Tensor(std)

    def forward(self, x):
        """
        Args:
            x (tensor batch): Input tensor.
        Returns:
            Normalized data.
        """
        return (x - self.mean.type_as(x)[None, :, None, None]) / self.std.type_as(x)[None, :, None, None]


class DivisiveNormalization2d(nn.Module):
    r"""Applies a 2D divisive normalization over an input signal composed of several input
    planes.

    In the simplest case, the output value of the layer with input size :math:`(N, C, H, W)`
    and output :math:`(N, C, H, W)`.

    Args:
        b_type: Type of suppressin field, must be one of (`linf`, `l1`, `l2`).
        b_size: The size of the suppression field, must be > 0.
        sigma: Constant added to suppression field, must be > 0.
        alpha: Global suppression rate, 0 means no suppression, 1 means complete suppression.

    Shape:
        - Input: :math:`(N, C, H, W)`
        - Output: :math:`(N, C, H, W)`

    Examples::

        >>> # suppression of size=3, sigma=1
        >>> d = DivisiveNormalization2d(b_size=3, sigma=1)
        >>> input = torch.randn(20, 16, 50, 50)
        >>> output = d(input)
    """

    def __init__(
            self,
            b_type: str = "linf",
            b_size: Union[int, Tuple[int, int]] = (3, 3),
            sigma: float = 1.0,
            alpha: float = 0.0,
            ) -> None:
        super(DivisiveNormalization2d, self).__init__()

        self.sigma = sigma
        self.alpha = alpha

        if isinstance(b_size, int):
            self.b_size = (b_size, b_size)
        else:
            self.b_size = b_size
        self.padding = (self.b_size[0]//2, self.b_size[1]//2)
        self.b_type = b_type

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.b_type == "linf":
            suppression_field = F.max_pool2d(
                torch.abs(input), self.b_size, 1, self.padding, 1)
        elif self.b_type == "l1":
            weight = torch.ones(
                (input.shape[1], 1, self.b_size[0], self.b_size[1])).to(input.device)
            suppression_field = F.conv2d(
                torch.abs(input), weight=weight, padding=self.padding, groups=input.shape[1])
        elif self.b_type == "l2":
            weight = torch.ones(
                (input.shape[1], 1, self.b_size[0], self.b_size[1])).to(input.device)
            suppression_field = torch.sqrt(F.conv2d(
                input**2, weight=weight, padding=self.padding, groups=input.shape[1]))
        else:
            raise NotImplementedError

        suppression_field = torch.mean(suppression_field, dim=1, keepdim=True)
        divisor = (self.sigma*torch.amax(suppression_field, dim=(2, 3),
                                         keepdim=True) + (1-self.sigma) * suppression_field + 0.0000001)
        return input / divisor

    def __repr__(self) -> str:
        s = "DivisiveNormalization2d("
        s += f'b_type={self.b_type}, b_size={self.b_size}, sigma={self.sigma}'
        if self.alpha > 0:
            s += f', alpha={self.alpha}'
        s += ")"
        return s


class AdaptiveThreshold(nn.Module):
    r"""
    Thresholds values x[x>threshold]
    Args:
        mean_scalar: threshold/means
    """

    def __init__(self, mean_scalar: float = 1.0) -> None:
        super(AdaptiveThreshold, self).__init__()

        self.mean_scalar = mean_scalar

    def _thresholding(self, x, threshold):
        return x*(x > threshold)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        means = input.mean(dim=(2, 3), keepdim=True)
        return self._thresholding(input, means*self.mean_scalar)

    def __repr__(self) -> str:
        s = f"AdaptiveThreshold(mean_scalar={self.mean_scalar})"
        return s


class FractionalThreshold(nn.Module):
    r"""
    FractionalThreshold 
    Args:
        remaining_ratio: Remaining activation ratio after Threshold.
    """

    def __init__(self, remaining_ratio: float = 0.1) -> None:
        super(FractionalThreshold, self).__init__()

        self.remaining_ratio = remaining_ratio

    def _thresholding(self, x: torch.Tensor, threshold: Union[torch.Tensor, float]) -> torch.Tensor:
        return x*(x > threshold)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        k = floor((1-self.remaining_ratio)*input.shape[2]*input.shape[3])
        thresholds = torch.kthvalue(input.view(
            *input.shape[:2], -1), k, dim=-1, keepdim=True)[0].unsqueeze(-1)
        return self._thresholding(input, thresholds)

    def __repr__(self) -> str:
        s = f"FractionalThreshold(remaining_ratio={self.remaining_ratio})"
        return s


class HaHBlock(nn.Module):
    r"""
    FractionalThreshold 
    Args:
        in_channels: Input channels for convolutional layer
        out_channels: Output channels for convolutional layer
        kernel_size: Kernel Size for convolutional layer
        padding: Padding for convolutional layer
        divisive_sigma: Hyperparameter for Divisive normalization
        ratio: Thresholding ratio (remaining_ratio)
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, padding=1, divisive_sigma=0.1, ratio=0.2) -> None:
        super(HaHBlock, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding
        self.divisive_sigma = divisive_sigma
        self.ratio = ratio

        self.conv_layer = ImplicitNormalizationConv(
            in_channels=self.in_channels, out_channels=self.out_channels, kernel_size=self.kernel_size, padding=self.padding, bias=False)
        self.relu = torch.nn.ReLU(inplace=False)
        self.normalization_layer = DivisiveNormalization2d(sigma=self.divisive_sigma)
        self.threshold_layer = FractionalThreshold(remaining_ratio=self.ratio)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        o = self.conv_layer(x)
        o = self.relu(o)
        o = self.normalization_layer(o)
        o = self.threshold_layer(o)
        return o

    def __repr__(self) -> str:
        s = f"HaHBlock(in_channels={self.in_channels}, out_channels={self.out_channels}, kernel_size={self.kernel_size}, padding={self.padding}, divisive_sigma={self.divisive_sigma}, ratio={self.ratio})"
        return s



from typing import Union, Tuple, Iterable

import torch
import torch.nn.functional as F
import torch.nn as nn

from ._layers import Normalize, HaHBlock

from math import floor


vgg_depth_dict = {
    'VGG11': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'VGG13': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'VGG16': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'VGG19': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}


class VGGBlock(nn.Module):
    r"""
    FractionalThreshold 
    Args:
        in_channels: Input channels for convolutional layer
        out_channels: Output channels for convolutional layer
        kernel_size: Kernel Size for convolutional layer
        padding: Padding for convolutional layer
        divisive_sigma: Hyperparameter for Divisive normalization
        ratio: Thresholding ratio (remaining_ratio)
    """

    def __init__(self, in_channels, out_channels, kernel_size=3, padding=1) -> None:
        super(VGGBlock, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.padding = padding

        self.conv_layer = torch.nn.Conv2d(
            in_channels=self.in_channels, out_channels=self.out_channels, kernel_size=self.kernel_size, padding=self.padding, bias=False)
        self.relu = torch.nn.ReLU(inplace=False)
        self.normalization_layer = nn.BatchNorm2d(self.out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        o = self.conv_layer(x)
        o = self.relu(o)
        o = self.normalization_layer(o)
        return o

    def __repr__(self) -> str:
        s = f"VGGBlock(in_channels={self.in_channels}, out_channels={self.out_channels}, kernel_size={self.kernel_size}, padding={self.padding})"
        return s


class HaHVGG16(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.normalize = Normalize((0.4914, 0.4822, 0.4465),
                                   (0.2471, 0.2435, 0.2616))
        self.vgg_depth = "VGG16"
        self.hah_layer_num = 7

        self.features = self.make_layers(vgg_depth_dict[self.vgg_depth])
        self.classifier = nn.Linear(512, 10)

        # self._initialize_weights()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.normalize(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x

    def _initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(
                    m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def make_layers(self, layer_list: Iterable[Union[str, int]]) -> nn.Sequential:
        blocks: list[nn.Module] = []
        in_channels = 3
        for layer_i, layer_info in enumerate(layer_list):
            if isinstance(layer_info, str):
                blocks += [nn.MaxPool2d(kernel_size=2, stride=2)]

            else:
                layer_width = layer_info

                if layer_i <= self.hah_layer_num:
                    blocks += [HaHBlock(in_channels=in_channels, out_channels=layer_width,
                                      kernel_size=3, padding=1, divisive_sigma=0.1, ratio=0.2)]
                else:
                    blocks += [VGGBlock(in_channels=in_channels, out_channels=layer_width,
                                      kernel_size=3, padding=1)]

                in_channels = layer_width

        return nn.Sequential(*blocks)

    @property
    def name(self) -> str:
        s = "HaHVGG16"
        return s


class VGG16(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.normalize = Normalize((0.4914, 0.4822, 0.4465),
                                   (0.2471, 0.2435, 0.2616))
        self.vgg_depth = "VGG16"
        self.hah_layer_num = 7

        self.features = self.make_layers(vgg_depth_dict[self.vgg_depth])
        self.classifier = nn.Linear(512, 10)

        # self._initialize_weights()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.normalize(x)
        x = self.features(x)
        x = torch.flatten(x, 1)
        x = self.classifier(x)
        return x

    def _initialize_weights(self) -> None:
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(
                    m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.constant_(m.bias, 0)

    def make_layers(self, layer_list: Iterable[Union[str, int]]) -> nn.Sequential:
        blocks: list[nn.Module] = []
        in_channels = 3
        for layer_i, layer_info in enumerate(layer_list):
            if isinstance(layer_info, str):
                blocks += [nn.MaxPool2d(kernel_size=2, stride=2)]

            else:
                layer_width = layer_info

                blocks = +[VGGBlock(in_channels=in_channels, out_channels=layer_width,
                                  kernel_size=3, padding=1)]

                in_channels = layer_width

        return nn.Sequential(*blocks)

    @property
    def name(self) -> str:
        s = "VGG16"
        return s
